@@ -9,7 +9,8 @@ import gi
 gi.require_version('Gtk', '3.0')
 from gi.repository import Gio, Gtk, GObject, Gdk, GdkPixbuf, GLib, Pango
 
-from GSettingsWidgets import *
+from SettingsWidgets import SidePage
+from xapp.GSettingsWidgets import *
 
 try:
     ENVIRON = os.environ['XDG_CURRENT_DESKTOP']
@@ -51,7 +52,7 @@ class Module:
     category = "prefs"
 
     def __init__(self, content_box):
-        keywords = _("startup, programs, boot, init, session")
+        keywords = _("startup, programs, boot, init, session, autostart, apps")
         sidePage = SidePage(_("Startup Applications"), "cs-startup-programs", keywords, content_box, module=self)
         self.sidePage = sidePage
 
@@ -383,16 +384,6 @@ class AutostartBox(Gtk.Box):
         main_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
         frame.add(main_box)
 
-        toolbar = Gtk.Toolbar.new()
-        Gtk.StyleContext.add_class(Gtk.Widget.get_style_context(toolbar), "cs-header")
-        label = Gtk.Label()
-        markup = GLib.markup_escape_text(title)
-        label.set_markup("<b>{}</b>".format(markup))
-        title_holder = Gtk.ToolItem()
-        title_holder.add(label)
-        toolbar.add(title_holder)
-        main_box.add(toolbar)
-
         scw = Gtk.ScrolledWindow()
         scw.expand = True
         scw.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
@@ -696,16 +687,19 @@ class AutostartRow(Gtk.ListBoxRow):
         self.delay_box.props.hexpand = False
         self.delay_box.set_margin_right(15)
         delay_time_markup = GLib.markup_escape_text(self.app.delay)
-        if delay_time_markup != "0":
-            label = Gtk.Label(_("Delay"))
-            self.delay_box.pack_start(label, False, False, 0)
-            self.delay_time_label = Gtk.Label()
-            self.delay_time_label.set_text(_("%s s") % delay_time_markup)
-            self.delay_time_label.get_style_context().add_class("dim-label")
-            self.delay_box.pack_start(self.delay_time_label, False, False, 0)
+
+        label = Gtk.Label(_("Delay"))
+        self.delay_box.pack_start(label, False, False, 0)
+        self.delay_time_label = Gtk.Label()
+        self.delay_time_label.set_markup(_("%s s") % delay_time_markup)
+        self.delay_time_label.get_style_context().add_class("dim-label")
+        self.delay_box.pack_start(self.delay_time_label, False, False, 0)
         grid.attach_next_to(self.delay_box, self.desc_box, Gtk.PositionType.RIGHT, 1, 1)
         self.delay_box.set_sensitive(app.enabled)
 
+        self.delay_box.show_all()
+        self.delay_box.set_no_show_all(True)
+        self.delay_box.set_visible(delay_time_markup != "0")
 
         switch_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
         switch_box.set_margin_top(5)
@@ -725,7 +719,8 @@ class AutostartRow(Gtk.ListBoxRow):
 
         self.name_label.set_markup("<b>{}</b>".format(name_markup))
         self.comment_label.set_markup("<small>{}</small>".format(comment_markup))
-        self.delay_time_label.set_markup(delay_time_markup)
+        self.delay_time_label.set_markup(_("%s s") % delay_time_markup)
+        self.delay_box.set_visible(delay_time_markup != "0")
 
     def on_switch_activated(self, switch, gparam):
         active = switch.get_active()
